package net.knarcraft.bookswithoutborders.command;

import net.knarcraft.bookswithoutborders.BooksWithoutBorders;
import org.bukkit.command.Command;
import org.bukkit.command.CommandExecutor;
import org.bukkit.command.CommandSender;
import org.bukkit.entity.Player;
import org.bukkit.inventory.ItemStack;

/**
 * Command executor for the give command
 */
public class CommandGive implements CommandExecutor {

    private final BooksWithoutBorders booksWithoutBorders;

    public CommandGive(BooksWithoutBorders booksWithoutBorders) {
        this.booksWithoutBorders = booksWithoutBorders;
    }

    @Override
    public boolean onCommand(CommandSender sender, Command command, String label, String[] args) {
        if (!(sender instanceof Player player)) {
            BooksWithoutBorders.sendErrorMessage(sender, "This command can only be used by a player!");
            return false;
        }

        if (args.length == 1 || args.length > 4) {
            BooksWithoutBorders.sendErrorMessage(player, "Incorrect number of arguments for this command!");
            return false;
        }

        if (args.length == 0) {
            BooksWithoutBorders.loadList.put(player.getName(), booksWithoutBorders.listFiles(player, false, false));
            return true;
        }

        String bookIdentifier = args[0];
        String receivingPlayerName = args[1];

        String copies = null;
        String isSigned = null;
        String copiesOrIsSigned = null;
        if (args.length == 4) {
            copies = args[2];
            isSigned = args[3];
        } else if (args.length == 3) {
            copiesOrIsSigned = args[2];
        }


        for (int x = 0; x < bookIdentifier.length(); x++) {
            if (!Character.isDigit(bookIdentifier.charAt(x))) {
                break;
            }
            if (x == bookIdentifier.length() - 1) {
                BooksWithoutBorders.loadList.put(player.getName(), booksWithoutBorders.listFiles(player, false, true));
            }
        }

        ItemStack newBook;
        Player receivingPlayer = booksWithoutBorders.getServer().getPlayer(receivingPlayerName);
        if (receivingPlayer == null) {
            BooksWithoutBorders.sendErrorMessage(player, "Player not found!");
            return false;
        }

        if (receivingPlayer.getInventory().firstEmpty() == -1) {
            BooksWithoutBorders.sendErrorMessage(player, "Receiving player must have space in their inventory to receive books!");
            return false;
        }

        //bwb give [bookname] [player] [numCopies] [issigned]
        try {

            if (isSigned != null && copies != null) {
                newBook = booksWithoutBorders.loadBook(player, booksWithoutBorders.cleanString(bookIdentifier), isSigned, "player", Integer.parseInt(copies));
            } else if (copiesOrIsSigned != null) {
                if (copiesOrIsSigned.equalsIgnoreCase("true") || copiesOrIsSigned.equalsIgnoreCase("false")) {
                    newBook = booksWithoutBorders.loadBook(player, booksWithoutBorders.cleanString(bookIdentifier), copiesOrIsSigned, "player");
                } else {
                    newBook = booksWithoutBorders.loadBook(player, booksWithoutBorders.cleanString(bookIdentifier), "true", "player", Integer.parseInt(copiesOrIsSigned));
                }
            } else {
                newBook = booksWithoutBorders.loadBook(player, booksWithoutBorders.cleanString(bookIdentifier), "true", "player");
            }

            if (newBook != null) {
                receivingPlayer.getInventory().addItem(newBook);
                BooksWithoutBorders.sendSuccessMessage(player, "Book sent!");
                BooksWithoutBorders.sendSuccessMessage(receivingPlayer, "Book received!");
                return true;
            } else {
                BooksWithoutBorders.sendErrorMessage(player, "Book failed to load!");
                return false;
            }
        } catch (NumberFormatException e) {
            BooksWithoutBorders.sendErrorMessage(player, "Invalid number of book copies specified!");
            return false;
        }
    }

}
