package net.knarcraft.bookswithoutborders.command;

import net.knarcraft.bookswithoutborders.BooksWithoutBorders;
import net.knarcraft.bookswithoutborders.utility.InventoryHelper;
import org.bukkit.ChatColor;
import org.bukkit.Material;
import org.bukkit.command.Command;
import org.bukkit.command.CommandExecutor;
import org.bukkit.command.CommandSender;
import org.bukkit.entity.Player;
import org.bukkit.inventory.ItemStack;
import org.bukkit.inventory.meta.ItemMeta;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * Command executor for the set lore command
 */
public class CommandSetLore implements CommandExecutor {

    @Override
    public boolean onCommand(CommandSender sender, Command command, String label, String[] args) {
        if (!(sender instanceof Player player)) {
            BooksWithoutBorders.sendErrorMessage(sender, "This command can only be used by a player!");
            return false;
        }

        if (args.length < 1) {
            BooksWithoutBorders.sendErrorMessage(player, "Missing a command argument!");
            return false;
        }

        ItemStack heldItem = InventoryHelper.getHeldItem(player, true);
        if (heldItem.getType() == Material.AIR) {
            BooksWithoutBorders.sendErrorMessage(player, "Must be holding an item to set lore!");
            return false;
        }

        //Treat all arguments as lore input
        StringBuilder rawLore = new StringBuilder();
        rawLore.append(args[0]);
        for (int x = 1; x < args.length; x++) {
            rawLore.append(" ").append(args[x]);
        }

        //Format lore
        rawLore = new StringBuilder(ChatColor.translateAlternateColorCodes('&', rawLore.toString()));
        String[] loreParts = rawLore.toString().split(BooksWithoutBorders.loreSeparator);
        List<String> newLore = new ArrayList<>(Arrays.asList(loreParts));

        //Update lore
        ItemMeta meta = heldItem.getItemMeta();
        if (meta == null) {
            BooksWithoutBorders.sendErrorMessage(player, "Could not get metadata from the held item!");
            return false;
        }
        meta.setLore(newLore);
        heldItem.setItemMeta(meta);
        BooksWithoutBorders.sendSuccessMessage(player, "Added lore to item!");
        return true;
    }

}
