package net.knarcraft.bookswithoutborders.utility;

import net.knarcraft.bookswithoutborders.BooksWithoutBorders;
import org.bukkit.ChatColor;
import org.bukkit.command.CommandSender;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

public class FileHelper {

    /**
     * Checks if a file path points to an actual book
     *
     * @param bookFile <p>The path to a book</p>
     * @return <p>True if the file exists and points to a book file</p>
     */
    public static boolean bookFileExists(String bookFile) {
        return ((new File(bookFile).isFile() && (bookFile.endsWith(".txt") ||
                bookFile.endsWith(".yml"))) || new File(bookFile + ".txt").isFile() ||
                new File(bookFile + ".yml").isFile()) && !bookFile.contains("../") && !bookFile.contains("..\\");
    }

    /**
     * Gets a book file given its path
     *
     * <p>This function adds extensions to the path as necessary, returning the first match.</p>
     *
     * @param bookPath <p>The path of the book to get</p>
     * @return <p>The file the path points to, or null otherwise</p>
     */
    public static File getBookFile(String bookPath) {
        if (!bookFileExists(bookPath)) {
            return null;
        }

        File bookFile = new File(bookPath);
        if (bookFile.exists()) {
            return bookFile;
        }

        File bookFileYml = new File(bookPath + ".yml");
        if (bookFileYml.exists()) {
            return bookFileYml;
        }

        File bookFileTxt = new File(bookPath + ".txt");
        if (bookFileTxt.exists()) {
            return bookFileTxt;
        }

        return null;
    }

    /**
     * Lists available files
     *
     * @param sender          <p>The command sender looking for files</p>
     * @param searchDirectory <p>The directory to search for files</p>
     * @param silent          <p>Whether to just return the list without printing it</p>
     * @return <p>A list of available files</p>
     */
    public static List<String> listFiles(CommandSender sender, File searchDirectory, boolean silent) {
        List<String> fileList = new ArrayList<>();
        File[] existingFiles = searchDirectory.listFiles();

        if (!searchDirectory.exists() || existingFiles == null || existingFiles.length == 0) {
            BooksWithoutBorders.sendErrorMessage(sender, "No books have been saved!");
            return null;
        }

        if (!silent) {
            BooksWithoutBorders.sendSuccessMessage(sender, "Available Books:");
        }
        for (File foundFile : existingFiles) {
            if (foundFile.isFile()) {
                fileList.add(foundFile.getName());
                if (!silent) {
                    sender.sendMessage(ChatColor.GRAY + "[" + fileList.size() + "] " + foundFile.getName());
                }
            }
        }

        return fileList;
    }

}
